<?php
/**
 * Provides a special function to invoke a page with dynamic parameters being passed
 *
 * @author Algorithmix (Gero Scholz) http://semeb.com/dpldemo

 * @version 0.1
 * @version 0.2
 *		added a help text if called without any arguments
 * @version 0.3
 *		different browsers add session variables in a different sequence; therefore we must check
 *		against "_session","UserName" and "UserID" simultaneously
 * @version 0.4
 *		added DebuG switch and recognition of "...Token"
 * @version 0.5
 *		changed algorithm to recoignize parameters
 */

class Call extends SpecialPage
{
        function Call() {
                SpecialPage::SpecialPage("Call");
                self::loadMessages();
        }

        
		function execute($par) {
		    global $wgOut, $wgRequest, $wgRawHtml, $wgUser;
		    $oldRawHtml = $wgRawHtml;
		    $wgRawHtml = false;         // disable raw html if it's enabled as this could be XSS security risk
            $this->setHeaders();
            		 
            global $_REQUEST;
       		$argkeys = array_keys($_REQUEST);
       		
       		// find the position of "title" and count succeeding arguments until we find one that matches
       		// one of the patterns which belong to typical session cookie variables
       		$argTitle=-1; $argCount=0; $n=0;
       		foreach ($argkeys as $argKey) {
	       		if ($argKey=='title') $argTitle = $n;
	       		else if ($argTitle>=0) {
		       		if (preg_match('/(UserName|UserID|_session|Token)$/',$argKey)) break;
		       		++$argCount;
	       		}
	       		$n++;
       		}	

       		$debug= !(strpos($wgRequest->getText('title'),'DebuG')===false);
       		if ($debug) {
		        $wgOut->addHtml("<pre>\n");
	       		foreach ($argkeys as $argKeyNr => $argKey) {
			        $wgOut->addHtml("$argKeyNr:$argKey:");
			        $wgOut->addHtml($wgRequest->getText($argKey)."\n",1);
	       		}
		        $wgOut->addHtml("\npar=$par\nargTitle=$argTitle\nargCount=$argCount</pre>");
       		}
       		
       		$wikitext=''; $n=0; $i=-1;
       		foreach ($argkeys as $argKeyNr => $argKey) {
	       		$i++;
	       		if ($i<$argTitle) continue;
	       		if ($i==$argTitle) {
		       		$wikitext .= preg_replace(',^[^/]+/,','',$wgRequest->getText($argKey),1);
		   			$wikitext = str_replace( ",", "|", $wikitext );
					$wikitext = str_replace( "_", " ", $wikitext );
		       		continue;
	       		}
	       		if (++$n > $argCount) break;
	       		$arg = $wgRequest->getText($argKey);
		       	if ($arg=='') {
			       	$arg = str_replace( "_", " ", $argKey );
					$wikitext .= ( '|' . $arg );
		       	} else {
			       	$arg = str_replace( "_", " ", $arg );
					$wikitext .= ( '|' . $argKey . '=' . $arg );
				}
       		}
       		
       		if ($wikitext=='' && $par!='') {
	       		// the first argument may contain parameters which are separated by comma
	       		// this is the case if [[Call,a=b]] syntax is used
	   			$wikitext = str_replace( ",", "|", $par );
				$wikitext = str_replace( "_", " ", $wikitext );
       		}
            
       		if ($wikitext=='' || $wikitext=='Special:Call' ) {
	       		$wgOut->addHtml("The Call extension expects a wiki page and optional parameters for that page as an argument.<br>\n<br>\n"
	       						."Example 1: &nbsp; <tt>[[Special:Call/My Template,parm1=value1]]</tt><br/>\n"
	       						."Example 2: &nbsp; <tt>[[Special:Call/Talk:My Discussion,parm1=value1]]</tt><br/>\n"
	       						."Example 3: &nbsp; <tt>[[Special:Call/:My Page,parm1=value1,parm2=value2]]</tt><br/>\n<br/>\n"
	       						."Example 4 (Browser URL): &nbsp; <tt>http://mydomain/mywiki/index.php?Special:Call/:My Page,parm1=value1</tt><br/>\n<br/>"
	       						."The <i>Call extension</i> will call the given page and pass the parameters.<br>\n"
	       						."You will see the contents of the called page and its title but its 'type' will be that of a special page,<br>"
	       						."i.e. such a page cannot be edited.<br>The contents you see may vary depending on the value of the parameters you passed.<br>\n<br>"
	       						."The <i>Call extension</i> is useful to build interactive applications with MediaWiki.<br>\n"
	       						."For an example see <a href='http://semeb.com/dpldemo/Template:Catlist'>the DPL GUI</a> ..<br/>\n"
	       						."In case of problems you can try <b>Special:Call/DebuG</b>\n");
       		}
       		else if ($debug) {
		        $wgOut->addHtml("<pre>\n{{".$wikitext."}}\n</pre>");
       		}
       		else {	       		
		        $wgOut->addHtml($wgOut->parse("{{".$wikitext."}}"));
	        }
            $skin = $wgUser->getSkin();
		 
		    $wgRawHtml = $oldRawHtml;

		    $newTitle = split("\,",$par,2);
		    if ($newTitle!=null && strlen($newTitle[0])>0) {
			    $newTitle[0]=str_replace("_"," ",$newTitle[0]);
			    $newTitle[0]=preg_replace("/^:/","",$newTitle[0]);
   				$wgOut->setPageTitle($newTitle[0]);
			}
  			// $wgOut->addMeta("http:expires", "0");
  
		}


        function loadMessages() {
                static $messagesLoaded = false;
                global $wgMessageCache;
                if ( $messagesLoaded ) return true;
                $messagesLoaded = true;

                require( dirname( __FILE__ ) . '/Call.i18n.php' );
                foreach ( $allMessages as $lang => $langMessages ) {
                        $wgMessageCache->addMessages( $langMessages, $lang );
                }
                return true;
        }
}
?>